@@ -4,6 +4,7 @@
 //! and general things in the operation of the Linux kernel.
 
 use std::cmp;
+use std::collections::HashSet;
 use std::str::FromStr;
 
 use bitflags::bitflags;
@@ -14,7 +15,7 @@ pub mod keys;
 pub mod random;
 
 /// Represents a kernel version, in major.minor.release version.
-#[derive(Debug, Copy, Clone, Eq, PartialEq)]
+#[derive(Debug, Copy, Clone, Eq, PartialEq, Hash)]
 pub struct Version {
     pub major: u8,
     pub minor: u8,
@@ -104,7 +105,151 @@ impl cmp::Ord for Version {
 
 impl cmp::PartialOrd for Version {
     fn partial_cmp(&self, other: &Self) -> Option<cmp::Ordering> {
-        Some(self.cmp(&other))
+        Some(self.cmp(other))
+    }
+}
+
+/// Represents a kernel type
+#[derive(Debug, Clone, Eq, PartialEq)]
+pub struct Type {
+    pub sysname: String,
+}
+
+impl Type {
+    pub fn new(sysname: String) -> Type {
+        Type { sysname }
+    }
+
+    /// Read the kernel type from current running kernel
+    ///
+    /// Defined in `include/linux/uts.h` as UTS_SYSNAME, default is "Linux".
+    /// The file is located at `/proc/sys/kernel/ostype`.
+    pub fn current() -> ProcResult<Self> {
+        read_value("/proc/sys/kernel/ostype")
+    }
+}
+
+impl FromStr for Type {
+    type Err = &'static str;
+
+    /// Parse a kernel type string
+    ///
+    /// Notice that in Linux source code, it is defined as a single string.
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        Ok(Type::new(s.to_string()))
+    }
+}
+
+/// Represents a kernel build information
+#[derive(Debug, Clone, Eq, PartialEq)]
+pub struct BuildInfo {
+    pub version: String,
+    pub flags: HashSet<String>,
+    /// This field contains any extra data from the /proc/sys/kernel/version file. It generally contains the build date of the kernel, but the format of the date can vary.
+    ///
+    /// A method named `extra_date` is provided which would try to parse some date formats. When the date format is not supported, an error will be returned. It depends on chrono feature.
+    pub extra: String,
+}
+
+impl BuildInfo {
+    pub fn new(version: &str, flags: HashSet<String>, extra: String) -> BuildInfo {
+        BuildInfo {
+            version: version.to_string(),
+            flags,
+            extra,
+        }
+    }
+
+    /// Read the kernel build information from current running kernel
+    ///
+    /// Generated by `scripts/mkcompile_h` when building the kernel.
+    /// The file is located at `/proc/sys/kernel/version`.
+    pub fn current() -> ProcResult<Self> {
+        read_value("/proc/sys/kernel/version")
+    }
+
+    /// Check if SMP is ON
+    pub fn smp(&self) -> bool {
+        self.flags.contains("SMP")
+    }
+
+    /// Check if PREEMPT is ON
+    pub fn preempt(&self) -> bool {
+        self.flags.contains("PREEMPT")
+    }
+
+    /// Check if PREEMPTRT is ON
+    pub fn preemptrt(&self) -> bool {
+        self.flags.contains("PREEMPTRT")
+    }
+
+    /// Return version number
+    ///
+    /// This would parse number from first digits of version string. For example, #21~1 to 21.
+    pub fn version_number(&self) -> ProcResult<u32> {
+        let mut version_str = String::new();
+        for c in self.version.chars() {
+            if c.is_ascii_digit() {
+                version_str.push(c);
+            } else {
+                break;
+            }
+        }
+        let version_number: u32 = version_str.parse().map_err(|_| "Failed to parse version number")?;
+        Ok(version_number)
+    }
+
+    /// Parse extra field to `DateTime` object
+    ///
+    /// This function may fail as TIMESTAMP can be various formats.
+    #[cfg(feature = "chrono")]
+    pub fn extra_date(&self) -> ProcResult<chrono::DateTime<chrono::Local>> {
+        if let Ok(dt) =
+            chrono::DateTime::parse_from_str(&format!("{} +0000", &self.extra), "%a %b %d %H:%M:%S UTC %Y %z")
+        {
+            return Ok(dt.with_timezone(&chrono::Local));
+        }
+        if let Ok(dt) = chrono::DateTime::parse_from_str(&self.extra, "%a, %d %b %Y %H:%M:%S %z") {
+            return Ok(dt.with_timezone(&chrono::Local));
+        }
+        Err(ProcError::Other("Failed to parse extra field to date".to_string()))
+    }
+}
+
+impl FromStr for BuildInfo {
+    type Err = &'static str;
+
+    /// Parse a kernel build information string
+    fn from_str(s: &str) -> Result<Self, Self::Err> {
+        let mut version = String::new();
+        let mut flags: HashSet<String> = HashSet::new();
+        let mut extra: String = String::new();
+
+        let mut splited = s.split(' ');
+        let version_str = splited.next();
+        if let Some(version_str) = version_str {
+            if let Some(stripped) = version_str.strip_prefix('#') {
+                version.push_str(stripped);
+            } else {
+                return Err("Failed to parse kernel build version");
+            }
+        } else {
+            return Err("Failed to parse kernel build version");
+        }
+
+        for s in &mut splited {
+            if s.chars().all(char::is_uppercase) {
+                flags.insert(s.to_string());
+            } else {
+                extra.push_str(s);
+                extra.push(' ');
+                break;
+            }
+        }
+        let remains: Vec<&str> = splited.collect();
+        extra.push_str(&remains.join(" "));
+
+        Ok(BuildInfo { version, flags, extra })
     }
 }
 
@@ -210,6 +355,7 @@ pub fn shmmni() -> ProcResult<u64> {
 
 bitflags! {
     /// Flags representing allowed sysrq functions
+    #[derive(Copy, Clone, Debug, Hash, Eq, PartialEq, PartialOrd, Ord)]
     pub struct AllowedFunctions : u16 {
         /// Enable control of console log level
         const ENABLE_CONTROL_LOG_LEVEL = 2;
@@ -248,7 +394,7 @@ impl SysRq {
         match self {
             SysRq::Disable => 0,
             SysRq::Enable => 1,
-            SysRq::AllowedFunctions(allowed) => allowed.bits,
+            SysRq::AllowedFunctions(allowed) => allowed.bits(),
         }
     }
 
@@ -302,7 +448,7 @@ pub fn threads_max() -> ProcResult<u32> {
 /// This function will return an error if that is not the case.
 pub fn set_threads_max(new_limit: u32) -> ProcResult<()> {
     if let Ok(kernel) = *KERNEL {
-        if kernel.major >= 4 && kernel.minor >= 1 && !(THREADS_MIN <= new_limit && new_limit <= THREADS_MAX) {
+        if kernel.major >= 4 && kernel.minor >= 1 && !(THREADS_MIN..=THREADS_MAX).contains(&new_limit) {
             return Err(ProcError::Other(format!(
                 "{} is outside the THREADS_MIN..=THREADS_MAX range",
                 new_limit
@@ -332,9 +478,60 @@ mod tests {
         assert_eq!(a, b);
     }
 
+    #[test]
+    fn test_type() {
+        let a = Type::from_str("Linux").unwrap();
+        assert_eq!(a.sysname, "Linux");
+    }
+
+    #[test]
+    fn test_build_info() {
+        // For Ubuntu, Manjaro, CentOS and others:
+        let a = BuildInfo::from_str("#1 SMP PREEMPT Thu Sep 30 15:29:01 UTC 2021").unwrap();
+        let mut flags: HashSet<String> = HashSet::new();
+        flags.insert("SMP".to_string());
+        flags.insert("PREEMPT".to_string());
+        assert_eq!(a.version, "1");
+        assert_eq!(a.version_number().unwrap(), 1);
+        assert_eq!(a.flags, flags);
+        assert!(a.smp());
+        assert!(a.preempt());
+        assert!(!a.preemptrt());
+        assert_eq!(a.extra, "Thu Sep 30 15:29:01 UTC 2021");
+        #[cfg(feature = "chrono")]
+        let _ = a.extra_date().unwrap();
+
+        // For Arch and others:
+        let b = BuildInfo::from_str("#1 SMP PREEMPT Fri, 12 Nov 2021 19:22:10 +0000").unwrap();
+        assert_eq!(b.version, "1");
+        assert_eq!(b.version_number().unwrap(), 1);
+        assert_eq!(b.flags, flags);
+        assert_eq!(b.extra, "Fri, 12 Nov 2021 19:22:10 +0000");
+        assert!(b.smp());
+        assert!(b.preempt());
+        assert!(!b.preemptrt());
+        #[cfg(feature = "chrono")]
+        let _ = b.extra_date().unwrap();
+
+        // For Debian and others:
+        let c = BuildInfo::from_str("#1 SMP Debian 5.10.46-4 (2021-08-03)").unwrap();
+        let mut flags: HashSet<String> = HashSet::new();
+        flags.insert("SMP".to_string());
+        assert_eq!(c.version, "1");
+        assert_eq!(c.version_number().unwrap(), 1);
+        assert_eq!(c.flags, flags);
+        assert_eq!(c.extra, "Debian 5.10.46-4 (2021-08-03)");
+        assert!(c.smp());
+        assert!(!c.preempt());
+        assert!(!c.preemptrt());
+        // Skip the date parsing for now
+    }
+
     #[test]
     fn test_current() {
         let _ = Version::current().unwrap();
+        let _ = Type::current().unwrap();
+        let _ = BuildInfo::current().unwrap();
     }
 
     #[test]
